(function ($, window, Drupal, once) {
  Drupal.webformScheduledTasksPlusFilters =
    Drupal.webformScheduledTasksPlusFilters ?? {};

  /**
   * Improve the UI of the rearrange filters dialog box.
   *
   * @constructor
   *
   * @param {jQuery} $table
   *   The table in the filter form.
   * @param {jQuery} $operator
   *   The filter groups operator element.
   */
  Drupal.webformScheduledTasksPlusFilters.RearrangeFilterHandler = function (
    $table,
    $operator,
  ) {
    /**
     * Keep a reference to the `<table>` being altered and to the div containing
     * the filter groups operator dropdown (if it exists).
     */
    this.table = $table;

    /**
     *
     * @type {jQuery}
     */
    this.operator = $operator;

    /**
     *
     * @type {boolean}
     */
    this.hasGroupOperator = this.operator.length > 0;

    /**
     * Keep a reference to all draggable rows within the table.
     *
     * @type {jQuery}
     */
    this.draggableRows = $table.find('.draggable');

    this.modifyTableDrag();
  };

  $.extend(
    Drupal.webformScheduledTasksPlusFilters.RearrangeFilterHandler.prototype,
    /** @lends Drupal.webformScheduledTasksPlusFilters.RearrangeFilterHandler# */ {
      /**
       * @method
       */
      modifyTableDrag() {
        const tableDrag =
          Drupal.tableDrag['webform-scheduled-tasks-plus-rearrange-filters'];
        const filterHandler = this;

        /**
         * Override the row.onSwap method from tabledrag.js.
         *
         * When a row is dragged to another place in the table, several things
         * need to occur.
         * - The row needs to be moved so that it's within one of the filter
         * groups.
         * - The operator cells that span multiple rows need their rowspan
         * attributes updated to reflect the number of rows in each group.
         * - The operator labels that are displayed next to each filter need to
         * be redrawn, to account for the row's new location.
         */
        tableDrag.row.prototype.onSwap = function () {
          if (filterHandler.hasGroupOperator) {
            // Make sure the row that just got moved (this.group) is inside one
            // of the filter groups (i.e. below an empty marker row or a
            // draggable). If it isn't, move it down one.
            const thisRow = $(this.group);
            const previousRow = thisRow.prev('tr');
            if (
              previousRow.length &&
              !previousRow.hasClass('group-message') &&
              !previousRow.hasClass('draggable')
            ) {
              // Move the dragged row down one.
              const next = thisRow.next();
              if (next[0].tagName === 'TR') {
                this.swap('after', next);
              }
            }
            filterHandler.updateRowspans();
          }
        };

        /**
         * Override the onDrop method from tabledrag.js.
         */
        tableDrag.onDrop = function () {
          // Make sure the "group" dropdown is properly updated when rows are
          // dragged into an empty filter group. This is borrowed heavily from
          // the block.js implementation of tableDrag.onDrop().
          const groupRow = $(this.rowObject.element)
            .prevAll('tr.group-message')
            .get(0);
          const groupName = groupRow.className.replace(
            /([^ ]+[ ]+)*group-([^ ]+)-message([ ]+[^ ]+)*/,
            '$2',
          );
          const groupField = $(
            'select.filter-group-select',
            this.rowObject.element,
          );
          if (!groupField[0].matches(`.filter-group-select-${groupName}`)) {
            const oldGroupName = groupField
              .attr('class')
              .replace(
                /([^ ]+[ ]+)*filter-group-select-([^ ]+)([ ]+[^ ]+)*/,
                '$2',
              );
            groupField
              .removeClass(`filter-group-select-${oldGroupName}`)
              .addClass(`filter-group-select-${groupName}`);
            groupField[0].value = groupName;
          }
        };
      },

      /**
       * Update the rowspan attribute of each cell containing an operator
       * dropdown.
       */
      updateRowspans() {
        let $row;
        let $currentEmptyRow;
        let draggableCount;
        let $operatorCell;
        const rows = $(this.table).find('tr');
        const length = rows.length;
        for (let i = 0; i < length; i++) {
          $row = $(rows[i]);
          if ($row.hasClass('group-title')) {
            // This row is a title row.
            // Keep a reference to the cell containing the dropdown operator.
            $operatorCell = $row.find('td.group-operator');
            // Assume this filter group is empty, until we find otherwise.
            draggableCount = 0;
            $currentEmptyRow = $row.next('tr');
            $currentEmptyRow
              .removeClass('group-populated')
              .addClass('group-empty');
            // The cell with the dropdown operator should span the title row and
            // the "this group is empty" row.
            $operatorCell.attr('rowspan', 2);
          } else if (
            $row.hasClass('draggable') &&
            Drupal.elementIsVisible(rows[i])
          ) {
            // We've found a visible filter row, so we now know the group isn't
            // empty.
            draggableCount++;
            $currentEmptyRow
              .removeClass('group-empty')
              .addClass('group-populated');
            // The operator cell should span all draggable rows, plus the title.
            $operatorCell.attr('rowspan', draggableCount + 1);
          }
        }
      },
    },
  );

  /**
   * Rearranges the filters.
   *
   * @type {Drupal~behavior}
   *
   * @prop {Drupal~behaviorAttach} attach
   *   Attach handlers to make it possible to rearrange the filters in the form
   *   in question.
   *   @see Drupal.webformScheduledTasksPlusFilters.RearrangeFilterHandler
   */
  Drupal.behaviors.webformScheduledTasksPlusFiltersRearrangeFilter = {
    attach(context) {
      // Only act on the rearrange filter form.
      if (
        typeof Drupal.tableDrag === 'undefined' ||
        typeof Drupal.tableDrag[
          'webform-scheduled-tasks-plus-rearrange-filters'
        ] === 'undefined'
      ) {
        return;
      }
      const table = once(
        'webform-scheduled-tasks-plus-rearrange-filters',
        '#webform-scheduled-tasks-plus-rearrange-filters',
        context,
      );
      const operator = once(
        'webform-scheduled-tasks-plus-rearrange-filters',
        '.js-form-item-filter-groups-operator',
        context,
      );
      if (table.length) {
        new Drupal.webformScheduledTasksPlusFilters.RearrangeFilterHandler(
          $(table),
          $(operator),
        );
      }
    },
  };
})(jQuery, window, Drupal, once);
