<?php

namespace Drupal\webform_scheduled_tasks_plus\Plugin\WebformScheduledTasks\ResultSet;

use Drupal\Component\Utility\SortArray;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\webform\WebformTokenManagerInterface;
use Drupal\webform_scheduled_tasks\Iterator\WebformIteratorAggregate;
use Drupal\webform_scheduled_tasks\Plugin\WebformScheduledTasks\ResultSetPluginBase;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Allow for custom filters to be set to export the results.
 *
 * @WebformScheduledResultSet(
 *   id = "custom_filters",
 *   label = @Translation("Custom Filters")
 * )
 */
final class CustomFilters extends ResultSetPluginBase {

  /**
   * Module Handler Service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * Database Connection Service.
   *
   * @var Drupal\Core\Database\Connection
   */
  protected Connection $connection;

  /**
   * Entity Type Manager Service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Webform Token Manager Service.
   *
   * @var \Drupal\webform\WebformTokenManagerInterface
   */
  protected WebformTokenManagerInterface $tokenManager;

  /**
   * Logger Service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->moduleHandler = $container->get('module_handler');
    $instance->connection = $container->get('database');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->tokenManager = $container->get('webform.token_manager');
    $instance->logger = $container->get('logger.factory')->get('webform_scheduled_tasks_plus');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getResultSet() {
    $this->initializeQueryDefaults();

    if (!empty($this->configuration['filters'])) {
      /** @var \Drupal\Core\Database\Query\SelectInterface $query */
      $query = $this->connection
        ->select('webform_submission_data');
      $query->fields('webform_submission_data', ['sid'])
        ->leftJoin(
          'webform_submission',
          NULL,
          'webform_submission_data.sid = webform_submission.sid'
        );
      $query->condition(
        'webform_submission.webform_id',
        $this->getScheduledTask()->getWebform()->id()
      );
      $groups = $this->combineGroups($this->configuration['filters']);

      $mainQueryGroup = $query->orConditionGroup();
      foreach ($groups as $subgroups) {
        $subQueryGroup = $query->andConditionGroup();
        foreach ($subgroups as $subgroup) {
          $conditionGroup = $subgroup['operator'] === 'AND' ?
            $query->andConditionGroup() :
            $query->orConditionGroup();
          foreach ($subgroup['filters'] as $filter) {
            preg_match('#(?P<table>\w+).(?P<field>\w+).?(?P<property>\w+)?#', $filter['field'], $parts);
            if (empty($parts) || !isset($parts['table']) || !isset($parts['field'])) {
              break;
            }
            $value = $this->alterValue($filter['value'], $filter['condition']);
            $condition = $this->getOperator($filter['condition']);
            if ($parts['table'] === 'webform_submission') {
              $conditionGroup->condition($parts['table'] . '.' . $parts['field'], $value, $condition);
            }
            else {
              $subGroup = $query->andConditionGroup();
              $subGroup->condition($parts['table'] . '.name', $parts['field']);
              $subGroup->condition($parts['table'] . '.value', $value, $condition);
              if (!empty($property)) {
                $subGroup->condition($parts['table'] . '.property', $parts['property']);
              }
              $conditionGroup->condition($subGroup);
            }
          }
          if ($conditionGroup->count() > 0) {
            $subQueryGroup->condition($conditionGroup);
          }
        }
        if ($subQueryGroup->count() > 0) {
          $mainQueryGroup->condition($subQueryGroup);
        }
      }

      // If there are any filters add them to the main.
      if ($mainQueryGroup->count() > 0) {
        $query->condition($mainQueryGroup);
      }

      $scheduledTask = $this->getScheduledTask();

      // Alter the query before hand.
      $this->moduleHandler->alter(
        [
          'webform_scheduled_tasks_plus_custom_filters_query',
          'webform_scheduled_tasks_plus_custom_filters_query_' . $scheduledTask->getWebform()->id(),
          'webform_scheduled_tasks_plus_custom_filters_query_' . $scheduledTask->id(),
        ],
        $query,
        $scheduledTask
      );

      try {
        $ids = $query->execute()->fetchCol();
        if (empty($ids)) {
          return [];
        }
      }
      catch (\Exception $e) {
        $this->logger->error($e->getMessage());
        return [];
      }
      $this->submissionQuery->condition('sid', $ids, 'IN');
    }

    return WebformIteratorAggregate::createFromQuery($this->submissionQuery)->getIterator();
  }

  /**
   * Combine the groups of data.
   *
   * The main purpose is to combine the groups and make writing the
   * query simpler. The idea is that AND conditions are evaluated together.
   *
   * In the example below:
   *
   * Group1 AND Group2 OR Group3 AND Group4
   *
   * The result array looks like:
   *
   * 0:
   *   - Group1
   *   - Group2
   * 1:
   *   - Group3
   *   - Group4
   *
   * In the example below:
   *
   * Group1 OR Group2 AND Group3 OR Group4
   *
   * The result array looks like:
   *
   * 0:
   *   - Group1
   * 1:
   *   - Group2
   *   - Group3
   * 2:
   *   - Group4
   *
   * @param array $groups
   *   Groups of data to combine.
   *
   * @return array
   *   Results of the combination.
   */
  protected function combineGroups(array $groups): array {
    $data = [];

    foreach ($groups as $group_id => $group) {
      // First key doesn't have a group_operator key so we add to the list.
      if (array_key_first($groups) === $group_id) {
        $data[][$group_id] = $group;
        continue;
      }

      // If the group_operator is AND add to the last array_key.
      if ($group['group_operator'] === 'AND') {
        $data[array_key_last($data)][$group_id] = $group;
      }
      else {
        // Otherwise append to the end of the array.
        $data[][$group_id] = $group;
      }
    }

    return $data;
  }

  /**
   * Alter the word condition to work with the query condition.
   *
   * @param string $condition
   *   Word Condition.
   *
   * @return string
   *   Query Condition that should be used.
   */
  protected function getOperator(string $condition): string {
    return match($condition) {
      'EQUAL', 'EMPTY' => '=',
      'NOT_EQUAL', 'NOT_EMPTY' => '<>',
      'CONTAINS', 'START_WITH', 'END_WITH' => 'LIKE',
      'NOT_CONTAINS', 'NOT_START_WITH', 'NOT_END_WITH' => 'NOT LIKE',
      'GREATER' => '>',
      'GREATER_EQUAL' => '>=',
      'LESS' => '<',
      'LESS_EQUAL' => '<=',
      'IN' => 'IN',
      'NOT_IN' => 'NOT IN',
      'NULL' => 'IS NULL',
      'NOT_NULL' => 'IS NOT NULL',
      default => '='
    };
  }

  /**
   * Based on the condition return the value and if it needs to be altered.
   *
   * @param string|array $value
   *   The value being checked against.
   * @param string $condition
   *   The condition being reviewed.
   *
   * @return string|array|null
   *   Return the modified string for the condition.
   */
  protected function alterValue(string|array $value, string $condition): string|array|null {
    return match($condition) {
      'CONTAINS', 'NOT_CONTAINS' => '%' . $value . '%',
      'START_WITH', 'NOT_START_WITH' => $value . '%',
      'END_WITH', 'NOT_END_WITH' => '%' . $value,
      'EMPTY', "NOT_EMPTY" => '',
      'IN', 'NOT_IN', 'BETWEEN' => (is_array($value) ? $value : explode('|', $value)),
      'NULL', 'NOT_NULL' => NULL,
      default => $value
    };
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'filters' => [],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['filter_wrapper'] = [
      '#type' => 'details',
      '#title' => $this->t('Filters'),
      '#open' => TRUE,
      '#prefix' => '<div id="filters-fieldset-wrapper">',
      '#suffix' => '</div>',
    ];

    $form['filter_wrapper']['filters'] = [
      '#type' => 'table',
      '#empty' => $this->t('No Filters Added'),
      '#tabledrag' => [
        [
          'action' => 'match',
          'relationship' => 'parent',
          'group' => 'filter-group-select',
          'subgroup' => 'filter-group-select',
          'source' => 'filter-id',
          'hidden' => TRUE,
        ],
        [
          'action' => 'order',
          'relationship' => 'sibling',
          'group' => 'filter-weight',
        ],
      ],
      '#attributes' => [
        'id' => 'webform-scheduled-tasks-plus-rearrange-filters',
      ],
    ];

    $groups = $this->configuration['filters'] ?? [];
    if ($form_state->get('groups') !== NULL) {
      $groups = $form_state->get('groups');
    }

    $group_options = [];
    array_walk($groups, function (&$item, $key) use (&$group_options) {
      $group_options[$key] = $key === 0 ? $this->t('Default Group') : $this->t('Group @group', ['@group' => $key + 1]);
    });
    foreach ($groups as $group_id => $group) {
      if (array_key_first($group_options) !== $group_id) {
        $form['filter_wrapper']['filters']['group_' . $group_id . '_operator'] = [
          [
            'group_operator' => [
              '#parents' => ['result_set_settings', 'group', $group_id, 'group_operator'],
              '#type' => 'select',
              '#title' => $this->t('Group Operator'),
              '#title_display' => 'invisible',
              '#options' => [
                'AND' => $this->t('AND'),
                'OR' => $this->t('OR'),
              ],
              '#default_value' => $group['group_operator'],
              '#wrapper_attributes' => [
                'class' => ['js-form-item-filter-groups-operator'],
              ],
            ],
            'description' => [
              '#markup' => '<div class="form-item__description"><span>' . $this->t('Either "group 0 AND group 1 AND group 2" or "group 0 OR group 1 OR group 2", etc') . '</span></div>',
            ],
            '#wrapper_attributes' => [
              'colspan' => 7,
            ],
          ],
          '#attributes' => [
            'class' => ['filter-group-operator-row'],
          ],
        ];
      }

      $form['filter_wrapper']['filters']['group_' . $group_id] = [
        'actions' => [
          'operator' => [
            '#parents' => ['result_set_settings', 'group', $group_id, 'operator'],
            '#type' => 'select',
            '#title' => $this->t('Operator'),
            '#options' => [
              'AND' => $this->t('AND'),
              'OR' => $this->t('OR'),
            ],
            '#default_value' => $group['operator'],
          ],
          '#wrapper_attributes' => [
            'rowspan' => max(2, count($group['filters']) + 1),
            'class' => ['group-operator'],
          ],
        ],
        'title' => [
          '#markup' => $group_id === 0 ? $this->t('Default Group') : $this->t('Group @group', ['@group' => $group_id + 1]),
          '#wrapper_attributes' => [
            'colspan' => 6,
            'class' => ['filter-group-title'],
          ],
        ],
        '#attributes' => [
          'class' => [
            'group-title',
            'group-title-' . $group_id,
          ],
        ],
      ];

      $form['filter_wrapper']['filters']['group_' . $group_id]['actions']['remove_group'] = [
        '#type' => 'submit',
        '#value' => $this->t('Remove Group'),
        '#submit' => [[static::class, 'removeFilter']],
        '#ajax' => [
          'callback' => [static::class, 'filterCallback'],
          'wrapper' => 'filters-fieldset-wrapper',
        ],
        '#name' => 'remove_group_' . $group_id,
        '#limit_validation_errors' => [
          ['result_set_settings', 'filters'],
          ['result_set_settings', 'filter_wrapper'],
          ['result_set_settings', 'group'],
        ],
        '#access' => $group_id > 0,
      ];

      $form['filter_wrapper']['filters']['group_' . $group_id . '_message'] = [
        'title' => [
          '#markup' => 'No filters have been added.',
          '#wrapper_attributes' => [
            'colspan' => 6,
          ],
        ],
        '#attributes' => [
          'class' => [
            'group-message',
            'group-' . $group_id . '-message',
            count($group['filters']) > 0 ? 'group-populated' : 'group-empty',
          ],
        ],
      ];

      foreach ($group['filters'] as $filter_id => $filter) {
        $row = [];

        $row['field'] = [
          '#type' => 'select',
          '#options' => $this->getWebformFilterFields(),
          '#title_display' => 'invisible',
          '#title' => $this->t('Field'),
          "#empty_option" => $this->t('- Select -'),
          '#default_value' => $filter['field'] ?? '',
          '#required' => TRUE,
        ];

        $row['condition'] = [
          '#type' => 'select',
          '#title_display' => 'invisible',
          '#title' => $this->t('Condition'),
          "#empty_option" => $this->t('- Select -'),
          '#default_value' => $filter['condition'] ?? '',
          '#required' => TRUE,
          '#options' => $this->getFilterConditions(),
        ];

        $row['value'] = [
          '#type' => 'textfield',
          '#title_display' => 'invisible',
          '#title' => $this->t('Value'),
          '#default_value' => $filter['value'] ?? '',
        ];

        $row['group'] = [
          '#type' => 'select',
          '#title' => $this->t('Group'),
          '#title_display' => 'invisible',
          '#options' => $group_options,
          '#default_value' => $group_id,
          '#attributes' => [
            'class' => [
              'filter-group-select',
              'filter-group-select-' . $group_id,
            ],
          ],
        ];

        $row['weight'] = [
          '#type' => 'weight',
          '#default_value' => $filter['weight'],
          '#delta' => 100,
          '#title_display' => 'invisible',
          '#title' => $this->t('Weight'),
          '#attributes' => [
            'class' => [
              'filter-weight',
              'filter-weight-' . $group_id,
            ],
          ],
        ];

        $row['remove'] = [
          'remove' => [
            '#type' => 'submit',
            '#value' => $this->t('Remove'),
            '#submit' => [[CustomFilters::class, 'removeFilter']],
            '#ajax' => [
              'callback' => [CustomFilters::class, 'filterCallback'],
              'wrapper' => 'filters-fieldset-wrapper',
            ],
            '#name' => 'remove_filter_' . $filter_id,
            '#limit_validation_errors' => [
              ['result_set_settings', 'filters'],
              ['result_set_settings', 'filter_wrapper'],
              ['result_set_settings', 'group'],
            ],
          ],
          'filter_id' => [
            '#type' => 'hidden',
            '#default_value' => $filter_id,
          ],
        ];

        $row['#attributes'] = [
          'class' => ['draggable', 'tabledrag-leaf'],
        ];

        $row['#parents'] = ['result_set_settings', 'filters', $filter_id];

        $form['filter_wrapper']['filters'][$filter_id] = $row;
      }
    }

    $form['filter_wrapper']['actions']['add_filter'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add filter'),
      '#submit' => [[static::class, 'addFilter']],
      '#ajax' => [
        'callback' => [static::class, 'filterCallback'],
        'wrapper' => 'filters-fieldset-wrapper',
      ],
      '#limit_validation_errors' => [
        ['result_set_settings', 'filters'],
        ['result_set_settings', 'filter_wrapper'],
        ['result_set_settings', 'group'],
      ],
      '#validate' => [],
    ];

    $form['filter_wrapper']['actions']['add_group'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add filter group'),
      '#submit' => [[CustomFilters::class, 'addFilterGroup']],
      '#ajax' => [
        'callback' => [CustomFilters::class, 'filterCallback'],
        'wrapper' => 'filters-fieldset-wrapper',
      ],
      '#limit_validation_errors' => [
        ['result_set_settings', 'filters'],
        ['result_set_settings', 'filter_wrapper'],
        ['result_set_settings', 'group'],
      ],
      '#validate' => [],
      '#access' => (count($groups) > 0 && count($groups[array_key_first($groups)]['filters']) > 0),
    ];

    $form['#attached']['library'][] = 'webform_scheduled_tasks_plus/custom_filters.filters';

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);
    $groups = CustomFilters::generateGroupData([
      'group' => $form_state->getValue('group'),
      'filters' => $form_state->getValue('filters'),
    ]);
    $form_state->set('groups', $groups);
  }

  /**
   * Combine the data to generate the group data for save.
   */
  protected static function generateGroupData(array $values = []): array {
    $groups = $values['group'] ?? [['operator' => 'AND', 'group_operator' => 'AND']];
    $filters = $values['filters'] ?? [];

    // Put the filters in their correct groups.
    foreach ($filters as $filter_id => $filter) {
      if (array_key_exists($filter['group'], $groups)) {
        if (!array_key_exists('filters', $groups[$filter['group']])) {
          $groups[$filter['group']]['filters'] = [];
        }

        $groups[$filter['group']]['filters'][$filter_id] = $filter;
      }
    }

    // Sort the filters by weight.
    foreach ($groups as &$group) {
      if (array_key_exists('filters', $group)) {
        uasort($group['filters'], function ($a, $b) {
          return SortArray::sortByWeightElement($a, $b);
        });
      }
      else {
        $group['filters'] = [];
      }
    }
    return $groups;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Remove the following values so they don't save to the config.
    $form_state->unsetValue('filter_wrapper');
    $form_state->unsetValue('filters');
    $form_state->unsetValue('group');
    parent::submitConfigurationForm($form, $form_state);
    $this->configuration['filters'] = $form_state->get('groups');
  }

  /**
   * Return the filter array key for callback.
   */
  public static function filterCallback(array &$form, FormStateInterface $form_state) {
    return $form['result_set_settings']['filter_wrapper'];
  }

  /**
   * Callback for adding filter row.
   */
  public static function addFilter(array &$form, FormStateInterface $form_state): void {
    $groups = $form_state->get('groups');
    $last_key = array_key_last($groups);
    $filter_id = static::generateRandomString(8);
    $groups[$last_key]['filters'][$filter_id] = [
      'field' => '',
      'condition' => 'EQ',
      'value' => '',
      'weight' => 0,
    ];
    $form_state->set('groups', $groups);
    $form_state->setRebuild();
  }

  /**
   * Callback for adding filter row.
   */
  public static function addFilterGroup(array &$form, FormStateInterface $form_state): void {
    $groups = $form_state->get('groups');
    $groups[] = [
      'operator' => 'AND',
      'group_operator' => 'AND',
      'filters' => [],
    ];
    $form_state->set('groups', $groups);
    $form_state->setRebuild();
  }

  /**
   * Callback for removing filter row and filter group.
   */
  public static function removeFilter(array &$form, FormStateInterface $form_state): void {
    $trigger = $form_state->getTriggeringElement();
    $groups = static::generateGroupData($form_state->getValue('result_set_settings'));
    if (str_starts_with($trigger['#name'], 'remove_filter_')) {
      $removeKey = (int) str_replace('remove_filter_', '', $trigger['#name']);
      foreach ($groups as &$group) {
        if (array_key_exists($removeKey, $group['filters'])) {
          unset($group['filters'][$removeKey]);
          break;
        }
      }
    }
    elseif (str_starts_with($trigger['#name'], 'remove_group_')) {
      $removeKey = (int) str_replace('remove_group_', '', $trigger['#name']);
      unset($groups[$removeKey]);
    }
    $form_state->set('groups', $groups);
    $form_state->setRebuild();
  }

  /**
   * Return a list of fields to filter on.
   */
  protected function getWebformFilterFields(): array {
    $webform = $this->getScheduledTask()->getWebform();

    $options = [];

    /** @var \Drupal\webform\WebformSubmissionStorageInterface $submission_storage */
    $submission_storage = $this->entityTypeManager->getStorage('webform_submission');
    $field_definitions = $submission_storage->getFieldDefinitions();
    $field_definitions = $submission_storage->checkFieldDefinitionAccess($webform, $field_definitions);
    foreach ($field_definitions as $key => $field_definition) {
      $option_key = sprintf('%s.%s', 'webform_submission', $key);
      $options[$option_key] = sprintf('%s (%s)', $field_definition['title'], $key);
    }

    $elements = $webform->getElementsInitializedFlattenedAndHasValue('view');
    // Replace tokens which can be used in an element's #title.
    $elements = $this->tokenManager->replace($elements, $webform);
    foreach ($elements as $key => $element) {
      $option_key = sprintf('%s.%s', 'webform_submission_data', $key);
      if (isset($element['#webform_composite_elements']) && is_array($element['#webform_composite_elements'])) {
        foreach ($element['#webform_composite_elements'] as $delta => $composite_element) {
          $options[$option_key . '.' . $delta] = sprintf(
            '%s - %s (%s)',
            $element['#admin_title'] ?: $element['#title'] ?: $key,
            $composite_element['#title'] ?: $delta,
            $delta
          );
        }
      }
      else {
        $options[$option_key] = sprintf('%s (%s)', $element['#admin_title'] ?: $element['#title'] ?: $key, $key);
      }
    }

    uasort($options, function ($a, $b) {
      return strcasecmp($a, $b);
    });
    return $options;
  }

  /**
   * Return a list of filter conditions.
   */
  protected function getFilterConditions(): array {
    $list = [
      'EQUAL' => $this->t('Equals'),
      'NOT_EQUAL' => $this->t('Not Equal'),
      'CONTAINS' => $this->t('Contains'),
      'NOT_CONTAINS' => $this->t('Not Contains'),
      'START_WITH' => $this->t('Starts With'),
      'NOT_START_WITH' => $this->t('Does Not Start With'),
      'END_WITH' => $this->t('Ends With'),
      'NOT_END_WITH' => $this->t('Does Not End With'),
      'BETWEEN' => $this->t('Between'),
      'GREATER' => $this->t('Greater Than'),
      'GREATER_EQUAL' => $this->t('Greater Than or Equal'),
      'LESS' => $this->t('Less Than'),
      'LESS_EQUAL' => $this->t('Less Than or Equal'),
      'IN' => $this->t('In List'),
      'NOT_IN' => $this->t('Not In List'),
      'EMPTY' => $this->t('Is Empty'),
      'NOT_EMPTY' => $this->t('Not Empty'),
      'NULL' => $this->t('Is Null'),
      'NOT_NULL' => $this->t('Is Not Null'),
    ];

    $this->moduleHandler->alter('webform_scheduled_tasks_plus_custom_filters_conditions', $list);

    return $list;
  }

  /**
   * Create a random string to use.
   *
   * @param int $length
   *   Length of the string to create.
   *
   * @return string
   *   Randomly generated string.
   */
  protected static function generateRandomString(int $length = 8): string {
    // Define the characters to use for the random string.
    $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $charactersLength = strlen($characters);
    $randomString = '';

    // Generate the random string.
    for ($i = 0; $i < $length; $i++) {
      $randomString .= $characters[rand(0, $charactersLength - 1)];
    }

    return $randomString;
  }

}
