<?php

namespace Drupal\webform_scheduled_tasks_plus\Plugin\WebformScheduledTasks\Task;

use Drupal\Core\Archiver\ArchiveTar;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\file\Entity\File;
use Drupal\file\FileInterface;
use Drupal\file\FileUsage\FileUsageInterface;
use Drupal\webform\Plugin\WebformExporterManagerInterface;
use Drupal\webform\WebformSubmissionExporterInterface;
use Drupal\webform_scheduled_tasks\Exception\HaltScheduledTaskException;
use Drupal\webform_scheduled_tasks\Plugin\WebformScheduledTasks\TaskPluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Creates an abstract class for file exporting.
 *
 * Used primarily as the base class for file exporting.
 */
abstract class FileExport extends TaskPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The submission exporter.
   *
   * @var \Drupal\webform\WebformSubmissionExporterInterface
   */
  protected $exporter;

  /**
   * The exporter manager.
   *
   * @var \Drupal\webform\Plugin\WebformExporterManagerInterface
   */
  protected $exporterManager;

  /**
   * The file usage service.
   *
   * @var \Drupal\file\FileUsage\FileUsageInterface
   */
  protected $fileUsage;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The file url generator service.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, WebformSubmissionExporterInterface $exporter, WebformExporterManagerInterface $exporterManager, FileUsageInterface $fileUsage, FileSystemInterface $fileSystem, FileUrlGeneratorInterface $fileUrlGenerator) {
    $this->exporter = $exporter;
    $this->exporterManager = $exporterManager;
    $this->fileUsage = $fileUsage;
    $this->fileSystem = $fileSystem;
    $this->fileUrlGenerator = $fileUrlGenerator;
    parent::__construct($configuration, $plugin_id, $plugin_definition, $this->messenger());
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    // @phpstan-ignore-next-line
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('webform_submission.exporter'),
      $container->get('plugin.manager.webform.exporter'),
      $container->get('file.usage'),
      $container->get('file_system'),
      $container->get('file_url_generator'),
    );
  }

  /**
   * Loop through the submissions and export them into a file.
   *
   * @param \Iterator $submissions
   *   List of submissions to loop through.
   * @param bool $returnFile
   *   If FALSE return the file entity otherwise return the string to the file.
   *
   * @return \Drupal\Core\File\FileInterface|null|string
   *   Return the created file, the location of the file, or null if successful.
   */
  protected function exportFile(\Iterator $submissions, bool $returnFile = TRUE): ?FileInterface {
    $exporter = $this->initializeExporter();

    $exporter_options = $exporter->getExportOptions();
    $exporter_options['exclude_columns'] = $this->configuration['exclude_fields'];
    $exporter->setWebformOptions($exporter_options);

    $exporter->writeHeader();
    $processed_submission_ids = [];
    foreach ($submissions as $submission) {
      $exporter->writeRecords([$submission]);
      $processed_submission_ids[] = $submission->id();
    }
    $exporter->writeFooter();

    // This is required to add actual submission data file into the associated
    // archive when files have been included inside the export.
    if ($exporter->getExporter()->isArchive()) {
      $exporter->writeExportToArchive();
    }

    // If no submissions were present while running this scheduled task, there
    // is no need to send an export.
    if (empty($processed_submission_ids)) {
      return NULL;
    }

    // Defensively catch errors in an export early to prevent any data loss,
    // check the file exists and the file has some data in it.
    if (!file_exists($this->getExportFileOrArchivePath($exporter)) || filesize($this->getExportFileOrArchivePath($exporter)) === 0) {
      throw new HaltScheduledTaskException('Export files are failing to generate or are empty.');
    }

    // Archive tar can throw warnings and produce a corrupt tar file without
    // ever throwing an exception, see d.o/project/drupal/issues/3026470.
    // Attempt to be a bit more defensive about this and try to verify a valid
    // archive was produced. With the php internal functions used to build the
    // archiver throwing warnings, a big risk is silent failure. The current
    // verification is a "best guess", given complexities around the number and
    // size of various files which should be in the archive.
    if ($exporter->getExporter()->isArchive() && !$this->verifyArchive($this->getExportFileOrArchivePath($exporter))) {
      throw new HaltScheduledTaskException('An invalid archive file was generated.');
    }

    $target_directory = $this->configuration['export_directory'];
    if (!$this->fileSystem->prepareDirectory($target_directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      throw new HaltScheduledTaskException('Could not create a directory for the exported files to be written to.');
    }
    $attempted_file_destination_uri = sprintf('%s/%s.%s',
      $this->configuration['export_directory'],
      $this->configuration['export_file_name'],
      $exporter->getExporter()->isArchive() ?
        $this->exporter->getExporter()->getArchiveFileExtension() :
        $this->exporter->getExporter()->getFileExtension()
    );
    if (!$unique_destination_file_uri = $this->fileSystem->copy($this->getExportFileOrArchivePath($exporter), $attempted_file_destination_uri)) {
      throw new HaltScheduledTaskException('Could not write the generated file to the private file system.');
    }

    if ($returnFile === FALSE) {
      return $attempted_file_destination_uri;
    }

    // Save a file entity with the data so the file can be deleted from the UI
    // if it has been downloaded.
    $file = File::create([
      'uri' => $unique_destination_file_uri,
      'filename' => $this->getExportFileOrArchiveName($exporter),
      'status' => FileInterface::STATUS_PERMANENT,
    ]);
    $file->save();

    return $file;
  }

  /**
   * {@inheritdoc}
   */
  public function executeTask(\Iterator $submissions) {
    $this->exportFile($submissions);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $exporter_options = $this->exporterManager->getOptions();
    return [
      'exclude_field' => [],
      'export_file_name' => '',
      'export_directory' => 'private://webform_scheduled_tasks_exports',
      'overwrite' => FALSE,
      'exporter_settings' => [],
      'exporter' => key($exporter_options),
      'delete_submissions' => FALSE,
      'include_attachments' => TRUE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['export_directory'] = [
      '#title' => $this->t('Export Directory'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $this->configuration['export_directory'],
      '#description' => $this->t('Location to export the file to. Directory must be writeable.'),
    ];

    $form['export_file_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('File name'),
      '#description' => $this->t('The name of the file to be exported.'),
      '#required' => TRUE,
      '#default_value' => $this->configuration['export_file_name'],
    ];

    $form['overwrite'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Overwrite existing file'),
      '#default_value' => $this->configuration['overwrite'],
      '#description' => $this->t('Overwrite existing file if it already exists.'),
    ];

    $form['delete_submissions'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Delete submissions after export'),
      '#default_value' => $this->configuration['delete_submissions'],
      '#description' => $this->t('Delete submissions after this task has been run.'),
    ];
    $form['include_attachments'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Include attachments'),
      '#default_value' => $this->configuration['include_attachments'],
      '#description' => $this->t('Include attachments uploaded by users in the exported archive.'),
    ];

    $this->buildExportPluginForm($form, $form_state);

    $form['exclude_fields'] = [
      '#title' => $this->t('Fields to be exported'),
      '#description' => $this->t('Include the following forms as part of the export.'),
      '#type' => 'webform_excluded_columns',
      '#webform_id' => $this->getScheduledTask()->getWebform()->id(),
      '#default_value' => $this->configuration['exclude_fields'],
    ];
    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * Build the export plugin form.
   */
  protected function buildExportPluginForm(&$form, FormStateInterface $form_state) {
    $form['exporter'] = [
      '#title' => $this->t('Export format'),
      '#type' => 'select',
      '#options' => $this->exporterManager->getOptions(),
      '#default_value' => $this->configuration['exporter'],
      '#ajax' => [
        'callback' => [static::class, 'ajaxCallback'],
        'wrapper' => 'exporter-settings',
      ],
    ];
    $chosen_exporter = $form_state->getValue('exporter', $this->configuration['exporter']);

    $plugin = $this->exporterManager->createInstance($chosen_exporter, $this->getConfiguration()['exporter_settings']);

    $form['exporter_settings']['exporter_settings'] = [
      '#prefix' => '<div id="exporter-settings">',
      '#suffix' => '</div>',
    ];
    $subform_state = SubformState::createForSubform($form['exporter_settings'], $form, $form_state);
    $form['exporter_settings'] += $plugin->buildConfigurationForm($form['exporter_settings'], $subform_state);
  }

  /**
   * An AJAX callback for the exporter_settings container.
   */
  public static function ajaxCallback($form, FormStateInterface $form_state) {
    return $form['task_settings']['exporter_settings'];
  }

  /**
   * Get an exporter with initialized settings.
   *
   * @return \Drupal\webform\WebformSubmissionExporterInterface
   *   The submissions exporter.
   */
  protected function initializeExporter() {
    // Set the scheduled task as the source entity so temporary files generated
    // will be with the context of the scheduled task. This ensures if someone
    // is generating an export from the UI, it does not use the same temporary
    // filenames.
    $this->exporter->setWebform($this->getScheduledTask()->getWebform());
    $this->exporter->setSourceEntity($this->getScheduledTask());
    $exporter_options = $this->configuration['exporter_settings'] + [
      'exporter' => $this->configuration['exporter'],
      'files' => $this->configuration['include_attachments'],
    ];
    $this->exporter->setExporter($exporter_options);
    return $this->exporter;
  }

  /**
   * Get the file or archive path, whichever is appropriate.
   *
   * @param \Drupal\webform\WebformSubmissionExporterInterface $initializedExporter
   *   The initialized exporter.
   *
   * @return string
   *   A path to an archive or file.
   */
  protected function getExportFileOrArchivePath(WebformSubmissionExporterInterface $initializedExporter) {
    if ($initializedExporter->getExporter()->isArchive()) {
      return $initializedExporter->getArchiveFilePath();
    }
    return $initializedExporter->getExportFilePath();
  }

  /**
   * Get the file or archive name, whichever is appropriate.
   *
   * @param \Drupal\webform\WebformSubmissionExporterInterface $initializedExporter
   *   The initialized exporter.
   *
   * @return string
   *   The name of a file.
   */
  protected function getExportFileOrArchiveName(WebformSubmissionExporterInterface $initializedExporter) {
    if ($initializedExporter->isArchive()) {
      return $initializedExporter->getArchiveFileName();
    }
    return $initializedExporter->getExportFileName();
  }

  /**
   * Verify a tar file.
   *
   * @param string $file
   *   The path to the archive.
   *
   * @return bool
   *   Check if an archive is empty.
   */
  protected function verifyArchive($file) {
    // Reading the contents of the archive is enough to raise an exception in
    // some cases and archives should always contain at least one file.
    $archive = new ArchiveTar($file);
    $content = $archive->listContent();
    return !empty($content);
  }

}
