<?php

namespace Drupal\webform_scheduled_tasks_plus\Plugin\WebformScheduledTasks\Task;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\file\FileUsage\FileUsageInterface;
use Drupal\webform\Plugin\WebformExporterManagerInterface;
use Drupal\webform\WebformSubmissionExporterInterface;
use Drupal\webform_scheduled_tasks_plus\WebformScheduledTasksTransportPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A task to export and transfer file to remote file system.
 *
 * @WebformScheduledTask(
 *   id = "remote_transfer",
 *   label = @Translation("Export and Transfer to External Location."),
 * )
 */
final class RemoteTransfer extends FileExport {

  /**
   * Transporter Plugin Manager.
   *
   * @var \Drupal\webform_scheduled_tasks_plus\WebformScheduledTasksTransportPluginManager
   */
  protected WebformScheduledTasksTransportPluginManager $transporterPluginManager;

  /**
   * Create a new RemoteTransfer object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, WebformSubmissionExporterInterface $exporter, WebformExporterManagerInterface $exporterManager, FileUsageInterface $fileUsage, FileSystemInterface $fileSystem, FileUrlGeneratorInterface $fileUrlGenerator, WebformScheduledTasksTransportPluginManager $transporterPluginManager) {
    $this->transporterPluginManager = $transporterPluginManager;
    parent::__construct($configuration, $plugin_id, $plugin_definition, $exporter, $exporterManager, $fileUsage, $fileSystem, $fileUrlGenerator);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('webform_submission.exporter'),
      $container->get('plugin.manager.webform.exporter'),
      $container->get('file.usage'),
      $container->get('file_system'),
      $container->get('file_url_generator'),
      $container->get('plugin.manager.webform_scheduled_tasks_plus_transport_plugin_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $transporter_options = $this->transporterPluginManager->getOptions();
    return [
      'transporter' => key($transporter_options),
      'transporter_settings' => [],
      'delete_after_transfer' => TRUE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $this->buildTransporterPluginForm($form, $form_state);

    return $form;
  }

  /**
   * Build the transporter plugin form.
   */
  protected function buildTransporterPluginForm(&$form, FormStateInterface $form_state) {
    $form['delete_after_transfer'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Delete after transfer'),
      '#default_value' => $this->configuration['delete_after_transfer'],
      '#description' => $this->t('Delete the saved file after it has been transferred.'),
    ];

    $form['transporter'] = [
      '#title' => $this->t('Transporter'),
      '#type' => 'select',
      '#required' => TRUE,
      '#description' => $this->t('Select the transporter to use.'),
      '#options' => $this->transporterPluginManager->getOptions(),
      '#default_value' => $this->configuration['transporter'],
      '#ajax' => [
        'callback' => [static::class, 'ajaxTransporterCallback'],
        'wrapper' => 'transporter-settings',
      ],
    ];
    $chosen_transporter = $form_state->getValue('transporter', $this->configuration['transporter']);
    $plugin = $this->transporterPluginManager->createInstance($chosen_transporter, $this->getConfiguration()['transporter_settings']);
    $form['transporter_settings'] = [
      '#prefix' => '<div id="transporter-settings">',
      '#suffix' => '</div>',
      'description' => [
        '#markup' => $plugin->description(),
      ],
    ];

    $subform_state = SubformState::createForSubform($form['transporter_settings'], $form, $form_state);
    $form['transporter_settings'] += $plugin->buildConfigurationForm($form['transporter_settings'], $subform_state);
  }

  /**
   * An AJAX callback for the exporter_settings container.
   */
  public static function ajaxTransporterCallback($form, FormStateInterface $form_state) {
    return $form['task_settings']['transporter_settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function executeTask(\Iterator $submissions) {
    $fileLocation = $this->exportFile($submissions, FALSE);

    /** @var \Drupal\webform_scheduled_tasks_plus\WebformScheduledTasksTransportPluginInterface $plugin */
    $plugin = $this->transporterPluginManager->createInstance(
      $this->getConfiguration()['transporter'],
      $this->getConfiguration()['transporter_settings']
    );

    $response = $plugin->transport($fileLocation);

    // If the transfer was successful. Delete after it has been transferred.
    if ($response && $this->configuration['delete_after_transfer']) {
      @unlink($fileLocation);
    }
  }

}
