<?php

namespace Drupal\webform_scheduled_tasks_plus\Plugin\WebformScheduledTasks\TransportPlugin;

use Drupal\Core\Form\FormStateInterface;
use Drupal\webform_scheduled_tasks_plus\FlysystemAdapterTrait;
use Drupal\webform_scheduled_tasks_plus\WebformScheduledTasksTransportPluginBase;
use League\Flysystem\FilesystemAdapter;
use League\Flysystem\Ftp\FtpAdapter;
use League\Flysystem\Ftp\FtpConnectionOptions;

/**
 * Create a FTP Transporter Plugin.
 *
 * @WebformScheduledTaskPlusTransportPlugin(
 *   id = "ftp",
 *   label = @Translation("FTP Connection"),
 *   description = @Translation("Transport via FTP connection."),
 * )
 */
final class FtpTransporter extends WebformScheduledTasksTransportPluginBase {

  use FlysystemAdapterTrait;

  /**
   * Return the adapter to use for transportation.
   */
  protected function getAdapter(): FilesystemAdapter {
    return new FtpAdapter(
      FtpConnectionOptions::fromArray([
        'host' => $this->configuration['host'],
        'root' => $this->configuration['path'],
        'username' => $this->configuration['username'],
        'password' => $this->configuration['password'],
        'port' => $this->configuration['port'],
        'ssl' => $this->configuration['ssl'],
        'timeout' => $this->configuration['timeout'],
        'utf8' => $this->configuration['utf8'],
        'passive' => $this->configuration['passive'],
      ])
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'host' => '',
      'port' => 21,
      'user' => '',
      'pass' => '',
      'path' => '',
      'ssl' => FALSE,
      'timeout' => 90,
      'utf8' => FALSE,
      'passive' => TRUE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['host'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Host Name'),
      '#default_value' => $this->configuration['host'],
      '#required' => TRUE,
      '#description' => $this->t('The hostname of the remote connection.'),
    ];

    $form['port'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Port Number'),
      '#default_value' => $this->configuration['port'],
      '#required' => TRUE,
      '#description' => $this->t('The port number of the remote connection.'),
    ];

    $form['user'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#default_value' => $this->configuration['user'],
      '#required' => TRUE,
      '#description' => $this->t('The username of the remote connection.'),
    ];

    $form['pass'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#default_value' => $this->configuration['pass'],
      '#description' => $this->t('The password of the remote connection.'),
    ];

    $form['path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Remote Path'),
      '#default_value' => $this->configuration['path'],
      '#required' => TRUE,
      '#description' => $this->t('The path of the remote connection.'),
    ];

    $form['ssl'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use SSL'),
      '#default_value' => $this->configuration['ssl'],
      '#description' => $this->t('If enabled, SSL will use the remote connection.'),
    ];

    $form['timeout'] = [
      '#type' => 'number',
      '#title' => $this->t('Timeout'),
      '#default_value' => $this->configuration['timeout'],
      '#min' => 1,
      '#description' => $this->t('The timeout in seconds.'),
      '#required' => TRUE,
    ];

    $form['passive'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Passive Mode'),
      '#default_value' => $this->configuration['passive'],
      '#description' => $this->t('If enabled, passive mode will use the remote connection.'),
    ];

    $form['utf8'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use UTF8'),
      '#default_value' => $this->configuration['utf8'],
      '#description' => $this->t('If enabled, utf8 will use the remote connection.'),
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

}
