<?php

namespace Drupal\webform_scheduled_tasks_plus\Plugin\WebformScheduledTasks\TransportPlugin;

use Drupal\Core\Form\FormStateInterface;
use Drupal\webform_scheduled_tasks_plus\FlysystemAdapterTrait;
use Drupal\webform_scheduled_tasks_plus\WebformScheduledTasksTransportPluginBase;
use League\Flysystem\FilesystemAdapter;
use League\Flysystem\PhpseclibV3\SftpAdapter;
use League\Flysystem\PhpseclibV3\SftpConnectionProvider;

/**
 * Create a SFTP Transporter Plugin.
 *
 * @WebformScheduledTaskPlusTransportPlugin(
 *   id = "sftp",
 *   label = @Translation("SFTP Connection"),
 *   description = @Translation("Transport via SFTP connection.")
 * )
 */
final class SftpTransporter extends WebformScheduledTasksTransportPluginBase {

  use FlysystemAdapterTrait;

  /**
   * Return the adapter to use for transportation.ub.
   */
  protected function getAdapter(): FilesystemAdapter {
    return new SftpAdapter(
      new SftpConnectionProvider(
        $this->configuration['host'],
        $this->configuration['username'],
        $this->configuration['password'],
        $this->configuration['private_key'],
        $this->configuration['private_key_pass'],
        $this->configuration['port'],
        $this->configuration['use_agent'],
        $this->configuration['timeout'],
        $this->configuration['max_tries'],
        $this->configuration['host_fingerprint']
      ),
      $this->configuration['path']
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'host' => '',
      'port' => 22,
      'user' => '',
      'pass' => '',
      'path' => '',
      'private_key' => '',
      'private_key_pass' => '',
      'use_agent' => FALSE,
      'timeout' => 10,
      'max_tries' => 4,
      'host_fingerprint' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['host'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Host Name'),
      '#default_value' => $this->configuration['host'],
      '#required' => TRUE,
      '#description' => $this->t('The hostname of the remote connection.'),
    ];

    $form['port'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Port Number'),
      '#default_value' => $this->configuration['port'],
      '#required' => TRUE,
      '#description' => $this->t('The port number of the remote connection.'),
    ];

    $form['user'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#default_value' => $this->configuration['user'],
      '#required' => TRUE,
      '#description' => $this->t('The username of the remote connection.'),
    ];

    $form['pass'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#default_value' => $this->configuration['pass'],
      '#description' => $this->t('The password of the remote connection.'),
    ];

    $form['path'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Remote Path'),
      '#default_value' => $this->configuration['path'],
      '#required' => TRUE,
      '#description' => $this->t('The path of the remote connection.'),
    ];

    $form['private_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Private Key'),
      '#default_value' => $this->configuration['private_key'],
      '#description' => $this->t('The private key for the remote connection. If this is used can leave Password field blank.'),
    ];

    $form['private_key_pass'] = [
      '#type' => 'password',
      '#title' => $this->t('Private Key Pass'),
      '#default_value' => $this->configuration['private_key_pass'],
      '#description' => $this->t('The private key for the remote connection.'),
    ];

    $form['use_agent'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use SSH Agent'),
      '#default_value' => $this->configuration['use_agent'],
      '#description' => $this->t('If this is used, it will use the ssh agent.'),
    ];

    $form['timeout'] = [
      '#type' => 'number',
      '#title' => $this->t('Timeout'),
      '#default_value' => $this->configuration['timeout'],
      '#description' => $this->t('The timeout in seconds.'),
      '#required' => TRUE,
      '#min' => 1,
    ];

    $form['max_tries'] = [
      '#type' => 'number',
      '#title' => $this->t('Max Tries'),
      '#default_value' => $this->configuration['max_tries'],
      '#description' => $this->t('The maximum number of times this connection is reached.'),
      '#required' => TRUE,
      '#min' => 1,
    ];

    $form['host_fingerprint'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Host Fingerprint'),
      '#default_value' => $this->configuration['host_fingerprint'],
      '#description' => $this->t('The host fingerprint of the remote connection.'),
    ];

    return parent::buildConfigurationForm($form, $form_state);
  }

}
